# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the definitions of arguments types for Spock (backend: attrs)"""

from enum import Enum, EnumMeta
from functools import partial
from typing import TypeVar, Union

import attr

from spock.backend.utils import _get_name_py_version
from spock.utils import _SpockGenericAlias, _SpockVariadicGenericAlias


class SavePath(str):
    """Spock special key for saving the Spock config to file

    Defines a special key use to save the current Spock config to file

    """

    def __new__(cls, x: str) -> str:
        return super().__new__(cls, x)


def _extract_base_type(typed):
    """Extracts the name of the type from a _GenericAlias

    Assumes that the derived types are only of length 1 as the __args__ are [0] recursed... this is not true for
    tuples

    Args:
        typed: the type of the parameter

    Returns:
        name of type
    """
    if hasattr(typed, "__args__") and not isinstance(typed, _SpockVariadicGenericAlias):
        name = _get_name_py_version(typed=typed)
        bracket_val = f"{name}[{_extract_base_type(typed.__args__[0])}]"
        return bracket_val
    else:
        bracket_value = _get_name_py_version(typed=typed)
    return bracket_value


def _recursive_generic_validator(typed):
    """Recursively assembles the validators for nested generic types

    Walks through the nested type structure and determines whether to recurse all the way to a base type. Once it
    hits the base type it bubbles up the correct validator that is nested within the upper validator

    Args:
        typed: input type

    Returns:
        return_type: recursively built deep_iterable validators

    """
    if hasattr(typed, "__args__") and not isinstance(typed, _SpockVariadicGenericAlias):
        # Iterate through since there might be multiple types?
        # Handle List and Tuple types
        if (
            _get_name_py_version(typed) == "List"
            or _get_name_py_version(typed) == "Tuple"
        ):
            # If there are more __args__ then we still need to recurse as it is still a GenericAlias
            if len(typed.__args__) > 1:
                return_type = attr.validators.deep_iterable(
                    member_validator=_recursive_generic_validator(typed.__args__),
                    iterable_validator=attr.validators.instance_of(typed.__origin__),
                )
            else:
                return_type = attr.validators.deep_iterable(
                    member_validator=_recursive_generic_validator(typed.__args__[0]),
                    iterable_validator=attr.validators.instance_of(typed.__origin__),
                )
            return return_type
        # Handle Dict types
        elif _get_name_py_version(typed) == "Dict":
            key_type, value_type = typed.__args__
            if key_type is not str:
                raise TypeError(
                    f"Unexpected key type of `{str(key_type.__name__)}` when attempting to handle "
                    f"GenericAlias type of Dict -- currently Spock only supports str as keys due "
                    f"to maintaining support for valid TOML and JSON files"
                )
            if hasattr(value_type, "__args__") and not isinstance(
                typed, _SpockVariadicGenericAlias
            ):
                return_type = attr.validators.deep_mapping(
                    value_validator=_recursive_generic_validator(value_type),
                    key_validator=attr.validators.instance_of(key_type),
                )
            else:
                return_type = attr.validators.deep_mapping(
                    value_validator=attr.validators.instance_of(value_type),
                    key_validator=attr.validators.instance_of(key_type),
                )
            return return_type
        else:
            raise TypeError(
                f"Unexpected type of `{str(typed)}` when attempting to handle GenericAlias types"
            )
    else:
        # If no more __args__ then we are to the base type and need to bubble up the type
        # But we need to check against base types and enums
        if isinstance(typed, EnumMeta):
            base_type, allowed = _check_enum_props(typed)
            return_type = attr.validators.and_(
                attr.validators.instance_of(base_type), attr.validators.in_(allowed)
            )
        else:
            return_type = attr.validators.instance_of(typed)
    return return_type


def _generic_alias_katra(typed, default=None, optional=False):
    """Private interface to create a subscripted generic_alias katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Handles: List[type], Tuple[type], Dict[type]

    Args:
        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:
        x: Attribute from attrs

    """
    # base python class from which a GenericAlias is derived
    base_typed = typed.__origin__
    if default is not None and optional:
        # if there's no default, but marked as optional, then set the default to None
        x = attr.ib(
            validator=attr.validators.optional(_recursive_generic_validator(typed)),
            type=base_typed,
            default=default,
            metadata={
                "optional": True,
                "base": _extract_base_type(typed),
                "type": typed,
            },
        )
    elif default is not None:
        x = attr.ib(
            validator=_recursive_generic_validator(typed),
            default=default,
            type=base_typed,
            metadata={"base": _extract_base_type(typed), "type": typed},
        )
        # x = attr.ib(validator=_recursive_generic_iterator(typed), default=default, type=base_typed,
        #             metadata={'base': _extract_base_type(typed)})
    elif optional:
        # if there's no default, but marked as optional, then set the default to None
        x = attr.ib(
            validator=attr.validators.optional(_recursive_generic_validator(typed)),
            type=base_typed,
            default=default,
            metadata={
                "optional": True,
                "base": _extract_base_type(typed),
                "type": typed,
            },
        )
    else:
        x = attr.ib(
            validator=_recursive_generic_validator(typed),
            type=base_typed,
            metadata={"base": _extract_base_type(typed), "type": typed},
        )
    return x


def _check_enum_props(typed):
    """Handles properties of enums

    Checks if all types of the enum are the same and assembles a list of allowed values

    Args:
        typed: the type of parameter (Enum)

    Returns:
        base_type: the base type of the Enum
        allowed: List of allowed values of the Enum

    """
    # First check if the types of Enum are the same
    type_set = {type(val.value) for val in typed}
    if len(type_set) > 1:
        raise TypeError(f"Enum cannot be defined with multiple types: {type_set}")
    base_type = list(type_set)[-1]
    allowed = [val.value for val in typed]
    return base_type, allowed


def _enum_katra(typed, default=None, optional=False):
    """Private interface to create a Enum typed katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Args:
        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:
        x: Attribute from attrs

    """
    # First check if the types of Enum are the same
    base_type, allowed = _check_enum_props(typed)
    if base_type.__name__ == "type":
        x = _enum_class_katra(
            typed=typed, allowed=allowed, default=default, optional=optional
        )
    else:
        x = _enum_base_katra(
            typed=typed,
            base_type=base_type,
            allowed=allowed,
            default=default,
            optional=optional,
        )
    return x


def _cast_enum_default(default):
    """Allows the enum default to be the specific value or the Enum structured value

    Checks if enum type and extracts the value from the Enum

    Args:
        default: the default value to assign if given

    Returns:
        default value or the Enum extracted value

    """
    if isinstance(default, Enum):
        return default.value
    else:
        return default


def _enum_base_katra(typed, base_type, allowed, default=None, optional=False):
    """Private interface to create a base Enum typed katra

    Here we handle the base types of enums that allows us to force a type check on the instance

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Args:
        typed: the type of the parameter to define
        base_type: underlying base type
        allowed: set of allowed values
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:
        x: Attribute from attrs

    """
    if default is not None and optional:
        x = attr.ib(
            validator=attr.validators.optional(
                [attr.validators.instance_of(base_type), attr.validators.in_(allowed)]
            ),
            default=_cast_enum_default(default),
            type=typed,
            metadata={"base": typed.__name__, "optional": True},
        )
    elif default is not None:
        x = attr.ib(
            validator=[
                attr.validators.instance_of(base_type),
                attr.validators.in_(allowed),
            ],
            default=_cast_enum_default(default),
            type=typed,
            metadata={"base": typed.__name__},
        )
    elif optional:
        x = attr.ib(
            validator=attr.validators.optional(
                [attr.validators.instance_of(base_type), attr.validators.in_(allowed)]
            ),
            default=_cast_enum_default(default),
            type=typed,
            metadata={"base": typed.__name__, "optional": True},
        )
    else:
        x = attr.ib(
            validator=[
                attr.validators.instance_of(base_type),
                attr.validators.in_(allowed),
            ],
            type=typed,
            metadata={"base": typed.__name__},
        )
    return x


def _in_type(instance, attribute, value, options):
    """attrs validator for class type enum

    Checks if the type of the class (e.g. value) is in the specified set of types provided. Also checks if the value
    is specified via the Enum definition

    Args:
        instance: current object instance
        attribute: current attribute instance
        value: current value trying to be set in the attrs instance
        options: list, tuple, or enum of allowed options

    Returns:
    """
    if type(value) not in options:
        raise ValueError(f"{attribute.name} must be in {options}")


def _enum_class_katra(typed, allowed, default=None, optional=False):
    """Private interface to create a base Enum typed katra

    Here we handle the class based types of enums. Seeing as these classes are generated dynamically we cannot
    force type checking of a specific instance however the in_ validator will catch an incorrect instance type

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Args:
        typed: the type of the parameter to define
        allowed: set of allowed values
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:
        x: Attribute from attrs

    """
    if default is not None and optional:
        x = attr.ib(
            validator=attr.validators.optional([partial(_in_type, options=allowed)]),
            default=_cast_enum_default(default),
            type=typed,
            metadata={"base": typed.__name__, "optional": True},
        )
    elif default is not None:
        x = attr.ib(
            validator=[partial(_in_type, options=allowed)],
            default=_cast_enum_default(default),
            type=typed,
            metadata={"base": typed.__name__},
        )
    elif optional:
        x = attr.ib(
            validator=attr.validators.optional([partial(_in_type, options=allowed)]),
            default=_cast_enum_default(default),
            type=typed,
            metadata={"base": typed.__name__, "optional": True},
        )
    else:
        x = attr.ib(
            validator=[partial(_in_type, options=allowed)],
            type=typed,
            metadata={"base": typed.__name__},
        )
    return x


def _type_katra(typed, default=None, optional=False):
    """Private interface to create a simple typed katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Handles: bool, string, float, int, List, and Tuple

    Args:
        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:
        x: Attribute from attrs

    """
    # Grab the name first based on if it is a base type or GenericAlias
    if isinstance(typed, type):
        name = typed.__name__
    elif isinstance(typed, _SpockGenericAlias):
        name = _get_name_py_version(typed=typed)
    else:
        raise TypeError("Encountered an unexpected type in _type_katra")
    special_key = None
    # Default booleans to false and optional due to the nature of a boolean
    if isinstance(typed, type) and name == "bool":
        optional = True
        # if it's a string -- it could be an env resolver -- pass it through
        if (not isinstance(default, str)) and (default is not True):
            default = False
    # For the save path type we need to swap the type back to it's base class (str)
    elif isinstance(typed, type) and name == "SavePath":
        optional = True
        special_key = name
        typed = str
    if default is not None and optional:
        # if a default is provided, that takes precedence
        x = attr.ib(
            validator=attr.validators.optional(attr.validators.instance_of(typed)),
            default=default,
            type=typed,
            metadata={"optional": True, "base": name, "special_key": special_key},
        )
    elif default is not None:
        # if a default is provided, that takes precedence
        x = attr.ib(
            validator=attr.validators.instance_of(typed),
            default=default,
            type=typed,
            metadata={"base": name, "special_key": special_key},
        )
    elif optional:
        x = attr.ib(
            validator=attr.validators.optional(attr.validators.instance_of(typed)),
            default=default,
            type=typed,
            metadata={"optional": True, "base": name, "special_key": special_key},
        )
    else:
        x = attr.ib(
            validator=attr.validators.instance_of(typed),
            type=typed,
            metadata={"base": name, "special_key": special_key},
        )
    return x


def _handle_optional_typing(typed):
    """Handles when a type hint is Optional

    Handles Optional[type] typing and strips out the base type to pass back to the creation of a katra which needs base
    typing

    Args:
        typed: type

    Returns:
        typed: type (modified if Optional)
        optional: boolean for katra creation
    """
    # Set optional to false
    optional = False
    # Check if it has __args__ to look for optionality as it is a GenericAlias -- also make sure it is not
    # callable as that also has __args__
    if hasattr(typed, "__args__") and not isinstance(typed, _SpockVariadicGenericAlias):
        # If it is more than one than it is most likely optional but check against NoneType in the tuple to verify
        # Check the length of type __args__
        type_args = typed.__args__
        # Optional[X] has type_args = (X, None) and is equal to Union[X, None]
        if (len(type_args) == 2) and (typed == Union[type_args[0], None]):
            typed = type_args[0]
            # Set the optional flag to true
            optional = True
    return typed, optional


def _callable_katra(typed, default=None, optional=False):
    """Private interface to create a Callable katra

    Here we handle the callable type katra that allows us to force a callable check on the value provided

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Handles: bool, string, float, int, List, and Tuple

    Args:
        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:
        x: Attribute from attrs

    """
    if default is not None and optional:
        x = attr.ib(
            validator=attr.validators.optional(attr.validators.is_callable()),
            default=default,
            type=typed,
            metadata={"optional": True, "base": _get_name_py_version(typed)},
        )
    elif default is not None:
        # if a default is provided, that takes precedence
        x = attr.ib(
            validator=attr.validators.is_callable(),
            default=default,
            type=typed,
            metadata={"base": _get_name_py_version(typed)},
        )
    elif optional:
        x = attr.ib(
            validator=attr.validators.optional(attr.validators.is_callable()),
            default=default,
            type=typed,
            metadata={"optional": True, "base": _get_name_py_version(typed)},
        )
    else:
        x = attr.ib(
            validator=attr.validators.is_callable(),
            type=typed,
            metadata={"base": _get_name_py_version(typed)},
        )
    return x


def katra(typed, default=None):
    """Public interface to create a katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Args:
        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:
        x: Attribute from attrs

    """
    # Handle optionals
    typed, optional = _handle_optional_typing(typed)
    if isinstance(typed, _SpockVariadicGenericAlias):
        x = _callable_katra(typed=typed, default=default, optional=optional)
    # We need to check if the type is a _GenericAlias so that we can handle subscripted general types
    # If it is subscript typed it will not be T which python uses as a generic type name
    elif isinstance(typed, _SpockGenericAlias) and (
        not isinstance(typed.__args__[0], TypeVar)
    ):
        x = _generic_alias_katra(typed=typed, default=default, optional=optional)
    elif isinstance(typed, EnumMeta):
        x = _enum_katra(typed=typed, default=default, optional=optional)
    else:
        x = _type_katra(typed=typed, default=default, optional=optional)
    # Add back in the OG type
    x.metadata.update({"og_type": typed})
    return x



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles mapping config arguments to a payload with both general and class specific sets"""


from typing import Any, Dict, Iterable

from _warnings import warn

from spock.exceptions import _SpockDuplicateArgumentError
from spock.graph import Graph


class SpockArguments:
    """Class that handles mapping the read parameter dictionary to general or class level arguments

    Attributes:
        _arguments: dictionary of arguments

    """

    def __init__(self, arguments: Dict, config_dag: Graph):
        """Init call for SpockArguments class

        Handles creating a clean arguments dictionary that can be cleanly mapped to spock classes

        Args:
            arguments: dictionary of parameters from the config file(s)
            config_dag: graph of the dependencies between spock classes

        """
        general_arguments = self._get_general_arguments(arguments, config_dag)
        attribute_name_to_config_name_mapping = (
            self._attribute_name_to_config_name_mapping(config_dag, general_arguments)
        )
        self._arguments = self._clean_arguments(arguments, general_arguments)
        self._assign_general_arguments_to_config(
            general_arguments, attribute_name_to_config_name_mapping
        )

    def __getitem__(self, key: str) -> Any:
        """Gets value from the _arguments dictionary

        Args:
            key: dictionary key

        Returns:
            argument at the specified key

        """
        return self._arguments[key]

    def __iter__(self) -> Iterable[str]:
        """Returns the next value of the keys within the _arguments dictionary

        Returns:
            current key for the _arguments dictionary

        """
        for key in self._arguments:
            yield key

    @property
    def items(self):
        """Returns the k,v tuple iterator for the _arguments dictionary"""
        return self._arguments.items()

    @property
    def keys(self):
        """Returns an iterator for the keys of the _arguments dictionary"""
        return self._arguments.keys()

    @property
    def values(self):
        """Returns an iterator for the values of the _arguments dictionary"""
        return self._arguments.values()

    @staticmethod
    def _get_general_arguments(arguments: Dict, config_dag: Graph) -> Dict:
        """Creates a dictionary of config file parameters that are defined at the general level (not class specific)

        Args:
            arguments: dictionary of parameters from the config file(s)
            config_dag: graph of the dependencies between spock classes

        Returns:
            dictionary of general level parameters

        """
        config_names = config_dag.node_names
        return {
            key: value
            for key, value in arguments.items()
            if key not in config_names and key != "config"
        }

    def _attribute_name_to_config_name_mapping(
        self, config_dag: Graph, general_arguments: Dict
    ) -> Dict:
        """Returns a mapping of names to spock config class parameter names

        Args:
            config_dag: graph of the dependencies between spock classes
            general_arguments: dictionary of arguments at the general level

        Returns:
            dictionary of parameters mapped to spock classes

        """
        attribute_name_to_config_name_mapping = {}
        for n in config_dag.nodes:
            for attr in n.__attrs_attrs__:
                if attr.name in general_arguments:
                    if self._is_duplicated_key(
                        attribute_name_to_config_name_mapping, attr.name, n.__name__
                    ):
                        raise _SpockDuplicateArgumentError(
                            f"`{attr.name}` key is located in more than one config and cannot be resolved automatically. "
                            f"Either specify the config name (`<config>.{attr.name}`) or change the key name in the config."
                        )
                    attribute_name_to_config_name_mapping[attr.name] = n.__name__

        return attribute_name_to_config_name_mapping

    @staticmethod
    def _is_duplicated_key(
        attribute_name_to_config_name_mapping: Dict, attr_name: str, config_name: str
    ) -> bool:
        """Checks if a duplicated key exists in multiple classes

        Args:
            attribute_name_to_config_name_mapping: dictionary of class specific name mappings
            attr_name:
            config_name:

        Returns:
            boolean if duplicated

        """
        return (
            attr_name in attribute_name_to_config_name_mapping
            and attribute_name_to_config_name_mapping[attr_name] != config_name
        )

    def _assign_general_arguments_to_config(
        self, general_arguments: Dict, attribute_name_to_config_name_mapping: Dict
    ) -> None:
        """Assigns the values from general definitions to values within specific classes if the specific definition
        doesn't exist

        Args:
            general_arguments:
            attribute_name_to_config_name_mapping:

        Returns:
            None
        """
        for arg, value in general_arguments.items():
            config_name = attribute_name_to_config_name_mapping[arg]
            if config_name in self._arguments:
                # Specific arguments supersede general arguments
                if arg not in self._arguments[config_name]:
                    self._arguments[config_name][arg] = value
                else:
                    warn(
                        f"Ignoring general argument `{arg}` for config `{config_name}`\n"
                        f"Specific argument value preceded general arguments.",
                        SyntaxWarning,
                    )
            else:
                self._arguments[config_name] = {arg: value}

    @staticmethod
    def _clean_arguments(arguments: Dict, general_arguments: Dict) -> Dict:
        """Sets up a clean dictionary for those not in the general dictionary

        Args:
            arguments: dictionary of all arguments
            general_arguments: dictionary of general level arguments

        Returns:
            clean dictionary of parameters not at the general level

        """
        clean_arguments = {}
        for arg, value in arguments.items():
            if arg not in general_arguments:
                clean_arguments[arg] = value
        return clean_arguments



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the tuner payload backend"""

from typing import Optional

from spock.backend.payload import BasePayload
from spock.backend.utils import _T, get_attr_fields


class TunerPayload(BasePayload):
    """Handles building the payload for tuners

    This class builds out the payload from config files of multiple types. It handles various
    file types and also composition of config files via a recursive calls

    Attributes:
        _loaders: maps of each file extension to the loader class

    """

    def __init__(self, s3_config: Optional[_T] = None):
        """Init for TunerPayload

        Args:
            s3_config: optional S3 config object

        """
        super().__init__(s3_config=s3_config)

    @staticmethod
    def _update_payload(base_payload, input_classes, ignore_classes, payload):
        # Get basic args
        attr_fields = get_attr_fields(input_classes=input_classes)
        # Get the ignore fields
        ignore_fields = get_attr_fields(input_classes=ignore_classes)
        for k, v in base_payload.items():
            if k not in ignore_fields:
                if k != "config":
                    # Dict infers that we are overriding a global setting in a specific config
                    if isinstance(v, dict):
                        # we're in a namespace
                        # Check for incorrect specific override of global def
                        if k not in attr_fields:
                            raise TypeError(
                                f"Referring to a class space `{k}` that is undefined"
                            )
                        for i_keys in v.keys():
                            if i_keys not in attr_fields[k]:
                                raise ValueError(
                                    f"Provided an unknown argument named `{k}.{i_keys}`"
                                )
                    if k in payload and isinstance(v, dict):
                        payload[k].update(v)
                    else:
                        payload[k] = v
                    # Handle tuple conversion here -- lazily
                    for ik, iv in v.items():
                        if "bounds" in iv:
                            iv["bounds"] = tuple(iv["bounds"])
        return payload

    @staticmethod
    def _handle_payload_override(payload, key, value):
        key_split = key.split(".")
        curr_ref = payload
        for idx, split in enumerate(key_split):
            # If the root isn't in the payload then it needs to be added but only for the first key split
            if idx == 0 and (split not in payload):
                payload.update({split: {}})
            # Check if it's the last value and figure out the override
            if idx == (len(key_split) - 1):
                # Handle bool(s) a bit differently as they are store_true
                if isinstance(curr_ref, dict) and isinstance(value, bool):
                    if value is not False:
                        curr_ref[split] = value
                # If we are at the dictionary level we should be able to just payload override
                elif isinstance(curr_ref, dict) and not isinstance(value, bool):
                    curr_ref[split] = value
                else:
                    raise ValueError(
                        f"cmd-line override failed for `{key}` -- "
                        f"Failed to find key `{split}` within lowest level Dict"
                    )
            # If it's not keep walking the current payload
            else:
                curr_ref = curr_ref[split]
        return payload



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Base handler Spock class"""

from abc import ABC
from typing import Optional

from spock.handlers import JSONHandler, TOMLHandler, YAMLHandler
from spock.utils import _T


class BaseHandler(ABC):
    """Base class for saver and payload

    Attributes:
        _writers: maps file extension to the correct i/o handler
        _s3_config: optional S3Config object to handle s3 access

    """

    def __init__(self, s3_config: Optional[_T] = None):
        self._supported_extensions = {
            ".yaml": YAMLHandler,
            ".toml": TOMLHandler,
            ".json": JSONHandler,
        }
        self._s3_config = s3_config

    def _check_extension(self, file_extension: str):
        if file_extension not in self._supported_extensions:
            raise TypeError(
                f"File extension {file_extension} not supported -- \n"
                f"File extension must be from {list(self._supported_extensions.keys())}"
            )



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""I/O handlers for various file formats"""

import json
import os
import re
from abc import ABC, abstractmethod
from pathlib import Path, PurePosixPath
from typing import ByteString, Dict, Optional, Tuple, Union
from warnings import warn

import pytomlpp
import yaml

from spock._version import get_versions
from spock.utils import _T, check_path_s3, path_object_to_s3path

__version__ = get_versions()["version"]


class Handler(ABC):
    """Base class for file type loaders

    ABC for loaders

    """

    def load(self, path: Path, s3_config: Optional[_T] = None) -> Dict:
        """Load function for file type

        This handles s3 path conversion for all handler types pre load call

        Args:
            path: path to file
            s3_config: optional s3 config object if using s3 storage

        Returns:
            dictionary of read file

        """
        path = self._handle_possible_s3_load_path(path=path, s3_config=s3_config)
        return self._post_process_config_paths(self._load(path=path))

    @staticmethod
    def _post_process_config_paths(payload):
        """
        Transform path string into path object
        """
        if (payload is not None) and "config" in payload:
            payload["config"] = [Path(c) for c in payload["config"]]
        return payload

    @abstractmethod
    def _load(self, path: str) -> Dict:
        """Private load function for file type

        Args:
            path: path to file

        Returns:
            dictionary of read file

        """
        raise NotImplementedError

    def _write_crypto(
        self,
        value: Union[str, ByteString],
        path: Path,
        name: str,
        crypto_name: str,
        create_path: bool,
        s3_config: Optional[_T],
    ):
        """Write values of the underlying cryptography data used to encode some spock values

        Args:
            value: current crypto attribute
            path: path to write out
            name: spock generated file name
            create_path: boolean to create the path if non-existent (for non S3)
            s3_config: optional s3 config object if using s3 storage
            crypto_name: name of the crypto attribute

        Returns:
            None

        """
        # Convert ByteString to str
        value = value.decode("utf-8") if isinstance(value, ByteString) else value
        write_path, is_s3 = self._handle_possible_s3_save_path(
            path=path, name=name, create_path=create_path, s3_config=s3_config
        )
        # We need to shim in the crypto value name into the name used for S3
        name_root, name_extension = os.path.splitext(name)
        name = f"{name_root}.{crypto_name}.yaml"
        # Also need to shim the crypto name into the full path
        root, extension = os.path.splitext(write_path)
        full_name = f"{root}.{crypto_name}.yaml"
        YAMLHandler.write({crypto_name: value}, full_name)
        # After write check if it needs to be pushed to S3
        if is_s3:
            self._check_s3_write(write_path, path, name, s3_config)

    def save(
        self,
        out_dict: Dict,
        info_dict: Optional[Dict],
        library_dict: Optional[Dict],
        path: Path,
        name: str,
        create_path: bool = False,
        s3_config: Optional[_T] = None,
        salt: Optional[str] = None,
        key: Optional[ByteString] = None,
    ):
        """Write function for file type

        This will handle local or s3 writes with the boolean is_s3 flag. If detected it will conditionally import
        the necessary addons to handle the upload

        Args:
            out_dict: payload to write
            info_dict: info payload to write
            library_dict: package info to write
            path: path to write out
            name: spock generated file name
            create_path: boolean to create the path if non-existent (for non S3)
            s3_config: optional s3 config object if using s3 storage
            salt: string of the salt used for crypto
            key: ByteString of the key used for crypto

        Returns:
        """
        write_path, is_s3 = self._handle_possible_s3_save_path(
            path=path, name=name, create_path=create_path, s3_config=s3_config
        )
        write_path = self._save(
            out_dict=out_dict,
            info_dict=info_dict,
            library_dict=library_dict,
            path=write_path,
        )
        # After write check if it needs to be pushed to S3
        if is_s3:
            self._check_s3_write(write_path, path, name, s3_config)
        # Write the crypto files if needed
        if (salt is not None) and (key is not None):
            # If the values are not none then write the salt and key into individual files
            self._write_crypto(salt, path, name, "salt", create_path, s3_config)
            self._write_crypto(key, path, name, "key", create_path, s3_config)

    @staticmethod
    def _check_s3_write(
        write_path: str, path: Path, name: str, s3_config: Optional[_T]
    ):
        """Handles writing to S3 if necessary

        Args:
            write_path: path the file was written to locally
            path: original path specified
            name: original file name
            s3_config: optional s3 config object if using s3 storage

        Returns:

        """
        try:
            from spock.addons.s3.utils import handle_s3_save_path

            handle_s3_save_path(
                temp_path=write_path,
                s3_path=str(PurePosixPath(path)),
                name=name,
                s3_config=s3_config,
            )
        except ImportError:
            print("Error importing spock s3 utils after detecting s3:// save path")

    @abstractmethod
    def _save(
        self,
        out_dict: Dict,
        info_dict: Optional[Dict],
        library_dict: Optional[Dict],
        path: str,
    ) -> str:
        """Write function for file type

        Args:
            out_dict: payload to write
            info_dict: info payload to write
            library_dict: package info to write
            path: path to write out

        Returns:
        """
        raise NotImplementedError

    @staticmethod
    def _handle_possible_s3_load_path(
        path: Path, s3_config: Optional[_T] = None
    ) -> Union[str, Path]:
        """Handles the possibility of having to handle loading from a S3 path

        Checks to see if it detects a S3 uri and if so triggers imports of s3 functionality and handles the file
        download

        Args:
            path: spock config path
            s3_config: optional s3 configuration object

        Returns:
            path: current path for the configuration file

        """
        is_s3 = check_path_s3(path=path)
        if is_s3:
            try:
                from spock.addons.s3.utils import handle_s3_load_path

                s3_path = path_object_to_s3path(path)
                path = handle_s3_load_path(path=s3_path, s3_config=s3_config)
            except ImportError:
                print("Error importing spock s3 utils after detecting s3:// load path")
        return path

    @staticmethod
    def _handle_possible_s3_save_path(
        path: Path, name: str, create_path: bool, s3_config: Optional[_T] = None
    ) -> Tuple[str, bool]:
        """Handles the possibility of having to save to a S3 path

        Checks to see if it detects a S3 uri and if so generates a tmp location to write the file to pre-upload

        Args:
            path: save path
            name: spock generated file name
            create_path: create the path for non s3 data
            s3_config: s3 config object

        Returns:
        """
        is_s3 = check_path_s3(path=path)
        if is_s3:
            if s3_config is None:
                raise ValueError(
                    "Save to S3 -- Missing S3Config object which is necessary to handle S3 style paths"
                )
            write_path = f"{s3_config.temp_folder}/{name}"
            # Strip double slashes if exist
            write_path = write_path.replace(r"//", r"/")
        else:
            # Handle the path logic for non S3
            if not os.path.exists(path) and create_path:
                os.makedirs(path)
            write_path = f"{path}/{name}"
        return write_path, is_s3

    @staticmethod
    def write_extra_info(
        path: str,
        info_dict: Dict,
        version: bool = True,
        write_mode: str = "w+",
        newlines: Optional[int] = None,
        header: Optional[str] = None,
    ):
        """Writes extra info to commented newlines

        Args:
            path: path to write out
            info_dict: info payload to write
            version: write the spock version string first
            write_mode: write mode for the file
            newlines: number of new lines to add to start

        Returns:
        """
        # Write the commented info as new lines
        with open(path, write_mode) as fid:
            if newlines is not None:
                for _ in range(newlines):
                    fid.write("\n")
            if header is not None:
                fid.write(header)
            # Write a spock header
            if version:
                fid.write(f"# Spock Version: {__version__}\n")
            # Write info dict if not None
            if info_dict is not None:
                for k, v in info_dict.items():
                    fid.write(f"{k}: {v}\n")
            fid.write("\n")


class YAMLHandler(Handler):
    """YAML class for loading YAML config files

    Base YAML class

    """

    # override default SafeLoader behavior to correctly
    # interpret 1e1 (as opposed to 1.e+1) as 10
    # https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number/30462009#30462009
    yaml.SafeLoader.add_implicit_resolver(
        "tag:yaml.org,2002:float",
        re.compile(
            """^(?:
         [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$""",
            re.X,
        ),
        list("-+0123456789."),
    )

    def _load(self, path: str) -> Dict:
        """YAML load function

        Args:
            path: path to YAML file

        Returns:
            base_payload: dictionary of read file

        """
        file_contents = open(path, "r").read()
        base_payload = yaml.safe_load(file_contents)
        return base_payload

    def _save(
        self,
        out_dict: Dict,
        info_dict: Optional[Dict],
        library_dict: Optional[Dict],
        path: str,
    ) -> str:
        """Write function for YAML type

        Args:
            out_dict: payload to write
            info_dict: info payload to write
            library_dict: package info to write
            path: path to write out
        Returns:
        """
        # First write the commented info
        self.write_extra_info(path=path, info_dict=info_dict)
        # Remove aliases in YAML dump
        yaml.Dumper.ignore_aliases = lambda *args: True
        self.write(out_dict, path)
        # Write the library info at the bottom
        self.write_extra_info(
            path=path,
            info_dict=library_dict,
            version=False,
            write_mode="a",
            newlines=2,
            header="################\n# Package Info #\n################\n",
        )
        return path

    @staticmethod
    def write(write_dict: Dict, path: str):
        # Remove aliases in YAML dump
        yaml.Dumper.ignore_aliases = lambda *args: True
        with open(path, "a") as yaml_fid:
            yaml.safe_dump(write_dict, yaml_fid, default_flow_style=False)


class TOMLHandler(Handler):
    """TOML class for loading TOML config files

    Base TOML class

    """

    def _load(self, path: str) -> Dict:
        """TOML load function

        Args:
            path: path to TOML file

        Returns:
            base_payload: dictionary of read file

        """
        base_payload = pytomlpp.load(path)
        return base_payload

    def _save(
        self,
        out_dict: Dict,
        info_dict: Optional[Dict],
        library_dict: Optional[Dict],
        path: str,
    ) -> str:
        """Write function for TOML type

        Args:
            out_dict: payload to write
            info_dict: info payload to write
            library_dict: package info to write
            path: path to write out

        Returns:
        """
        # First write the commented info
        self.write_extra_info(path=path, info_dict=info_dict)
        with open(path, "a") as toml_fid:
            pytomlpp.dump(out_dict, toml_fid)
        # Write the library info at the bottom
        self.write_extra_info(
            path=path, info_dict=library_dict, version=False, write_mode="a", newlines=2
        )
        return path


class JSONHandler(Handler):
    """JSON class for loading JSON config files

    Base JSON class

    """

    def _load(self, path: str) -> Dict:
        """JSON load function

        Args:
            path: path to JSON file

        Returns:
            base_payload: dictionary of read file

        """
        with open(path) as json_fid:
            base_payload = json.load(json_fid)
        return base_payload

    def _save(
        self,
        out_dict: Dict,
        info_dict: Optional[Dict],
        library_dict: Optional[Dict],
        path: str,
    ) -> str:
        """Write function for JSON type

        Args:
            out_dict: payload to write
            info_dict: info payload to write
            library_dict: package info to write
            path: path to write out

        Returns:
        """
        if (info_dict is not None) or (library_dict is not None):
            warn(
                "JSON does not support comments and thus cannot save extra info to file... removing extra info"
            )
        with open(path, "a") as json_fid:
            json.dump(out_dict, json_fid, indent=4, separators=(",", ": "))
        return path



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the tuner builder backend"""

from spock.backend.builder import BaseBuilder
from spock.utils import make_argument


class TunerBuilder(BaseBuilder):
    def __init__(self, *args, **kwargs):
        """TunerBuilder init

        Args:
            *args: list of input classes that link to a backend
            configs: None or List of configs to read from
            desc: description for the arg parser
            no_cmd_line: flag to force no command line reads
            **kwargs: any extra keyword args
        """
        super().__init__(*args, module_name="spock.addons.tune.config", **kwargs)

    @staticmethod
    def _make_group_override_parser(parser, class_obj, class_name):
        """Makes a name specific override parser for a given class obj

        Takes a class object of the backend and adds a new argument group with argument names given with name
        Class.val.(unrolled config parameters) so that individual parameters specific to a class can be overridden.

        Args:
            parser: argument parser
            class_obj: instance of a backend class
            class_name: used for module matching

        Returns:
            parser: argument parser with new class specific overrides

        """
        attr_name = class_obj.__name__
        group_parser = parser.add_argument_group(
            title=str(attr_name) + " Specific Overrides"
        )
        for val in class_obj.__attrs_attrs__:
            val_type = val.metadata["type"] if "type" in val.metadata else val.type
            for arg in val_type.__attrs_attrs__:
                arg_name = f"--{str(attr_name)}.{val.name}.{arg.name}"
                group_parser = make_argument(arg_name, arg.type, group_parser)
        return parser

    def _extract_fnc(self, val, module_name):
        return self._extract_other_types(val.type, module_name)



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Attr utility functions for Spock"""

import importlib
from typing import Any, Callable, Dict, List, Tuple, Type, Union

from cryptography.fernet import Fernet

from spock.exceptions import _SpockValueError
from spock.utils import _C, _T, _SpockVariadicGenericAlias


def encrypt_value(value, key, salt):
    # Make the class to encrypt
    encrypt = Fernet(key=key)
    # Encrypt the plaintext value
    salted_password = value + salt
    # encode to utf-8 -> encrypt -> decode from utf-8
    return encrypt.encrypt(str.encode(salted_password)).decode()


def decrypt_value(value, key, salt):
    # Make the class to encrypt
    decrypt = Fernet(key=key)
    # Decrypt back to plaintext value
    salted_password = decrypt.decrypt(str.encode(value)).decode()
    return salted_password[: -len(salt)]


def _str_2_callable(val: str, **kwargs):
    """Tries to convert a string representation of a module and callable to the reference to the callable

    If the module of callable cannot be found on the PYTHONPATH then an exception is raised

    Args:
        val: string rep of a module and callable
        **kwargs: in case additional keyword args are passed

    Returns:
        reference to a callable

    Raises:
        _SpockValueError

    """
    str_field = str(val)
    module, fn = str_field.rsplit(".", 1)
    try:
        call_ref = getattr(importlib.import_module(module), fn)
    except Exception as e:
        raise _SpockValueError(
            f"Attempted to import module {module} and callable {fn} however it could not be found on the current "
            f"python path: {e}"
        )
    return call_ref


def _callable_2_str(val: Callable, **kwargs):
    """Converts a callable to a str based on the module and name

    Args:
        val: callable object
        **kwargs: in case additional keyword args are passed

    Returns:
        string of module.name

    """
    return f"{val.__module__}.{val.__name__}"


def _recurse_callables(val: _T, fnc: Callable, check_type: Type = str):
    """Recurses through objects casting any callables to the correct format

    Handles both strings to callables and callables to strings depending on the function passed in

    Args:
        val: current object
        fnc: callable
        check_type: type to check against to verify conversion

    Returns:
        object with mapped callables/strings

    """
    if isinstance(val, (list, List, tuple, Tuple)):
        out = []
        for v in val:
            if isinstance(val, (list, List, tuple, Tuple, dict, Dict)):
                out.append(_recurse_callables(v, fnc, check_type))
            else:
                out.append(fnc(v))
        out = type(val)(out)
    elif isinstance(val, (dict, Dict)):
        out = {}
        for k, v in val.items():
            if isinstance(val, (list, List, tuple, Tuple, dict, Dict)):
                out.update({k: _recurse_callables(v, fnc, check_type)})
            else:
                out.update({k: fnc(v)})
    elif isinstance(val, check_type):
        out = fnc(val)
    # Fall back to just passing the unmodified value back
    else:
        out = val
    return out


def _get_name_py_version(typed: _T):
    """Gets the name of the type depending on the python version

    Args:
        typed: the type of the parameter

    Returns:
        name of the type

    """
    return typed._name if hasattr(typed, "_name") else typed.__name__


def get_attr_fields(input_classes: List):
    """Gets the attribute fields from all classes

    Args:
        input_classes: current list of input classes

    Returns:
        dictionary of all attrs attribute fields

    """
    return {
        attr.__name__: [val.name for val in attr.__attrs_attrs__]
        for attr in input_classes
    }


def get_type_fields(input_classes: List):
    """Creates a dictionary of names and types

    Args:
        input_classes: list of input classes

    Returns:
        type_fields: dictionary of names and types

    """
    # Parse out the types if generic
    type_fields = {}
    for attr in input_classes:
        input_attr = {}
        for val in attr.__attrs_attrs__:
            if "type" in val.metadata:
                input_attr.update({val.name: val.metadata["type"]})
            else:
                input_attr.update({val.name: None})
        type_fields.update({attr.__name__: input_attr})
    return type_fields


def flatten_type_dict(type_dict: Dict):
    """Flattens a nested dictionary

    Args:
        type_dict: dictionary of types that are generic

    Returns:
        flat_dict: flatten dictionary to a single level

    """
    flat_dict = {}
    for k, v in type_dict.items():
        if isinstance(v, dict):
            # return_dict = flatten_type_dict(v, input_dict)
            return_dict = flatten_type_dict(
                v,
            )
            flat_dict.update(return_dict)
        else:
            flat_dict[k] = v
    return flat_dict


def _get_iter(value: Union[List, Dict]):
    """Returns the iterator for the type

    Args:
        value: instance of type List or Dict

    Returns:
        iterator

    """
    if isinstance(value, (list, List)):
        return enumerate(value)
    elif isinstance(value, (dict, Dict)):
        return value.items()
    else:
        raise ValueError(f"Cannot get iterator for type `{type(value)}`")


def convert_to_tuples(
    input_dict: Dict, base_type_dict: Dict, flat_type_dict: Dict, class_names: List
):
    """Convert lists to tuples

    Payloads from markup come in as Lists and not Tuples. This function turns lists in to tuples for the payloads
    so the attr values are set correctly. Will call itself recursively if it is iterable. Handles list of classes
    specifically

    Args:
        input_dict: input dictionary
        base_type_dict: dictionary of nested types
        flat_type_dict: dictionary of names with generic types
        class_names: List of base spock class names

    Returns:
        updated_dict: a dictionary with lists converted to tuples

    """
    updated_dict = {}
    for k, v in input_dict.items():
        if k != "config":
            # If v is any iterable we need to step into the nested structure
            if isinstance(v, (dict, Dict, list, List, tuple, Tuple)):
                # We've hit a fundamental declared type -- recurse via type checking
                if k in flat_type_dict:
                    updated = _recursive_list_to_tuple(
                        k, v, flat_type_dict[k], class_names
                    )
                    if updated:
                        updated_dict.update({k: updated})
                # Have to handle this specifically -- this is lists of spock classes
                # here we need to update a list instead of the dict directly
                elif isinstance(v, (list, List)) and k in class_names:
                    updated_list = []
                    # Iterate over list of classes
                    for _, val in _get_iter(v):
                        updated = convert_to_tuples(
                            val, base_type_dict, flat_type_dict, class_names
                        )
                        updated_list.append(updated)
                    updated_dict.update({k: updated_list})
                # Keep recursing through the structure
                else:
                    updated = convert_to_tuples(
                        v, base_type_dict, flat_type_dict, class_names
                    )
                    if updated:
                        updated_dict.update({k: updated})
            # If there is no nested structure then just map
            else:
                updated_dict.update({k: v})
    return updated_dict


def deep_update(source: Dict, updates: Dict):
    """Deeply updates a dictionary

    Iterates through a dictionary recursively to update individual values within a possibly nested dictionary
    of dictionaries

    Args:
        source: source dictionary
        updates: updates to the dictionary

    Returns:
        source: updated version of the source dictionary

    """
    for k, v in updates.items():
        if isinstance(v, (dict, Dict)) and v:
            updated_dict = deep_update(source.get(k), v)
            if updated_dict:
                source[k] = updated_dict
        else:
            source[k] = v
    return source


def _recursive_list_to_tuple(key: str, value: Any, typed: _T, class_names: List):
    """Recursively turn lists into tuples

    Recursively looks through a pair of value and type and sets any of the possibly nested type of value to tuple
    if tuple is the specified type

    Args:
        key: name of parameter
        value: value to check and set type if necessary
        typed: type of the generic alias to check against
        class_names: list of all spock class names

    Returns:
        value: updated value with correct type casts

    """
    # Check for __args__ as it signifies a generic and make sure it's not already been cast as a tuple
    # from a composed payload
    if (
        hasattr(typed, "__args__")
        and not isinstance(value, tuple)
        and not (isinstance(value, str) and value in class_names)
        and not isinstance(typed, _SpockVariadicGenericAlias)
    ):
        # Force those with origin tuple types to be of the defined length
        # Here check for tuple and check length
        if (typed.__origin__ in (tuple, Tuple)) and len(value) != len(typed.__args__):
            raise ValueError(
                f"Tuple(s) are of a defined length -- For parameter {key} the length of the provided argument "
                f"({len(value)}) does not match the length of the defined argument ({len(typed.__args__)})"
            )
        # need to recurse before casting as we can't set values in a tuple with idx
        # Since it's generic it should be iterable  -- thus recurse and check its children
        if isinstance(value, (list, List)):
            # Then we continue to recurse
            # Iterate the list type and recurse call
            for idx, val in _get_iter(value):
                value[idx] = _recursive_list_to_tuple(
                    key, val, typed.__args__[0], class_names
                )
        elif isinstance(value, (dict, Dict)):
            # Iterate the dict type and recurse call
            for k, v in _get_iter(value):
                key_type, val_type = typed.__args__
                value[k] = _recursive_list_to_tuple(key, v, val_type, class_names)
        # We need to catch the mismatch if the current type is list and the given type is Tuple
        # First check if list and then swap to tuple if the origin is tuple -- only do this on the bubble up since
        # we can't assign by idx once we change to Tuple
        if isinstance(value, (list, List)) and typed.__origin__ in (tuple, Tuple):
            value = tuple(value)
    else:
        return value
    return value



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""
import argparse
from abc import ABC, abstractmethod
from enum import EnumMeta
from typing import ByteString, Dict, List

import attr

from spock.args import SpockArguments
from spock.backend.field_handlers import RegisterSpockCls
from spock.backend.help import attrs_help
from spock.backend.spaces import BuilderSpace
from spock.backend.wrappers import Spockspace
from spock.graph import Graph
from spock.utils import _C, _T, _SpockVariadicGenericAlias, make_argument


class BaseBuilder(ABC):  # pylint: disable=too-few-public-methods
    """Base class for building the backend specific builders

    This class handles the interface to the backend with the generic ConfigArgBuilder so that different
    backends can be used to handle processing

    Attributes

        _input_classes: list of input classes that link to a backend
        _graph: Graph, graph of the dependencies between spock classes
        _max_indent: maximum to indent between help prints
        _module_name: module name to register in the spock module space
        save_path: list of path(s) to save the configs to
        _lazy: attempts to lazily find @spock decorated classes registered within sys.modules["spock"].backend.config
        _salt: salt use for crypto purposes
        _key: key used for crypto purposes

    """

    def __init__(
        self,
        *args,
        max_indent: int = 4,
        module_name: str,
        lazy: bool,
        salt: str,
        key: ByteString,
        **kwargs,
    ):
        """Init call for BaseBuilder

        Args:
            *args: iterable of @spock decorated classes
            max_indent: max indent for pretty print of help
            module_name: module name to register in the spock module space
            lazy: lazily find @spock decorated classes
            salt: cryptographic salt
            key: cryptographic key
            **kwargs: keyword args
        """
        self._input_classes = args
        self._lazy = lazy
        self._salt = salt
        self._key = key
        self._graph = Graph(input_classes=self.input_classes, lazy=self._lazy)
        # Make sure the input classes are updated -- lazy evaluation
        self._input_classes = self._graph.nodes
        self._module_name = module_name
        self._max_indent = max_indent
        self.save_path = None

    @property
    def input_classes(self):
        """Returns the graph of dependencies between spock classes"""
        return self._input_classes

    @property
    def dag(self):
        """Returns the underlying graph DAG"""
        return self._graph.dag

    @property
    def graph(self):
        """Returns the underlying graph object"""
        return self._graph

    @staticmethod
    @abstractmethod
    def _make_group_override_parser(
        parser: argparse.ArgumentParser, class_obj: _C, class_name: str
    ) -> argparse.ArgumentParser:
        """Makes a name specific override parser for a given class obj

        Takes a class object of the backend and adds a new argument group with argument names given with name
        Class.name so that individual parameters specific to a class can be overridden.

        Args:
            parser: argument parser
            class_obj: instance of a backend class
            class_name: used for module matching

        Returns:
            parser: argument parser with new class specific overrides

        """

    def handle_help_info(self) -> None:
        """Handles walking through classes to get help info

        For each class this function will search __doc__ and attempt to pull out help information for both the class
        itself and each attribute within the class

        Returns:
            None

        """
        attrs_help(
            input_classes=self.input_classes,
            module_name=self._module_name,
            extract_fnc=self._extract_fnc,
            max_indent=self._max_indent,
        )

    def generate(self, dict_args: Dict) -> Spockspace:
        """Method to auto-generate the actual class instances from the generated args

        Based on the generated arguments groups and the args read in from the config file(s)
        this function instantiates the classes with the necessary field or attr values

        Args:
            dict_args: dictionary of arguments from the configs

        Returns:
            namespace containing automatically generated instances of the classes
        """
        spock_space_kwargs = self.resolve_spock_space_kwargs(self._graph, dict_args)
        return Spockspace(**spock_space_kwargs)

    def resolve_spock_space_kwargs(self, graph: Graph, dict_args: Dict) -> Dict:
        """Build the dictionary that will define the spock space.

        Args:
            graph: Dependency graph of nested spock configurations
            dict_args: dictionary of arguments from the configs

        Returns:
            dictionary containing automatically generated instances of the classes
        """
        # Empty dictionary that will be mapped to a SpockSpace via spock classes
        spock_space = {}
        # Assemble the arguments dictionary and BuilderSpace
        builder_space = BuilderSpace(
            arguments=SpockArguments(dict_args, graph), spock_space=spock_space
        )
        # For each root recursively step through the definitions
        for spock_cls in graph.roots:
            # Initial call to the RegisterSpockCls generate function (which will handle recursing if needed)
            spock_instance, special_keys = RegisterSpockCls.recurse_generate(
                spock_cls, builder_space, self._salt, self._key
            )
            builder_space.spock_space[spock_cls.__name__] = spock_instance

            for special_key, value in special_keys.items():
                setattr(self, special_key, value)

        return spock_space

    def build_override_parsers(
        self, parser: argparse.ArgumentParser
    ) -> argparse.ArgumentParser:
        """Creates parsers for command-line overrides

        Builds the basic command line parser for configs and help then iterates through each attr instance to make
        namespace specific cmd line override parsers

        Args:
            parser: argument parser

        Returns:
            parser: argument parser with new class specific overrides

        """
        # Build out each class override specific parser
        for val in self.input_classes:
            parser = self._make_group_override_parser(
                parser=parser, class_obj=val, class_name=self._module_name
            )
        return parser

    def _extract_other_types(self, typed: _T, module_name: str) -> List:
        """Takes a high level type and recursively extracts any enum or class types

        Args:
            typed: highest level type
            module_name: name of module to match

        Returns:
            return_list: list of nums (dot notation of module_path.enum_name or module_path.class_name)

        """
        return_list = []
        if hasattr(typed, "__args__") and not isinstance(
            typed, _SpockVariadicGenericAlias
        ):
            for val in typed.__args__:
                recurse_return = self._extract_other_types(val, module_name)
                if isinstance(recurse_return, list):
                    return_list.extend(recurse_return)
                else:
                    return_list.append(self._extract_other_types(val, module_name))
        elif isinstance(typed, EnumMeta) or (typed.__module__ == module_name):
            return [f"{typed.__module__}.{typed.__name__}"]
        return return_list

    @abstractmethod
    def _extract_fnc(self, val, module_name):
        """Function that gets the nested lists within classes

        Args:
            val: current attr
            module_name: matching module name

        Returns:
            list of any nested classes/enums

        """


class AttrBuilder(BaseBuilder):
    """Attr specific builder

    Class that handles building for the attr backend

    Attributes

        input_classes: list of input classes that link to a backend
        _configs: None or List of configs to read from
        _create_save_path: boolean to make the path to save to
        _desc: description for the arg parser
        _no_cmd_line: flag to force no command line reads
        save_path: list of path(s) to save the configs to

    """

    def __init__(self, *args, **kwargs):
        """AttrBuilder init

        Args:
            *args: list of input classes that link to a backend
            **kwargs: any extra keyword args
        """
        super().__init__(*args, module_name="spock.backend.config", **kwargs)

    @staticmethod
    def _make_group_override_parser(
        parser: argparse.ArgumentParser, class_obj: _C, class_name: str
    ) -> argparse.ArgumentParser:
        """Makes a name specific override parser for a given class obj

        Takes a class object of the backend and adds a new argument group with argument names given with name
        Class.name so that individual parameters specific to a class can be overridden.

        Args:
            parser: argument parser
            class_obj: instance of a backend class
            class_name: used for module matching

        Returns:
            parser: argument parser with new class specific overrides

        """
        attr_name = class_obj.__name__
        group_parser = parser.add_argument_group(
            title=str(attr_name) + " Specific Overrides"
        )
        for val in class_obj.__attrs_attrs__:
            val_type = val.metadata["type"] if "type" in val.metadata else val.type
            # Check if the val type has __args__ -- this catches GenericAlias classes
            # TODO (ncilfone): Fix up this super super ugly logic
            if (
                not isinstance(val_type, _SpockVariadicGenericAlias)
                and hasattr(val_type, "__args__")
                and ((list(set(val_type.__args__))[0]).__module__ == class_name)
                and attr.has((list(set(val_type.__args__))[0]))
            ):
                args = list(set(val_type.__args__))[0]
                for inner_val in args.__attrs_attrs__:
                    arg_name = f"--{str(attr_name)}.{val.name}.{args.__name__}.{inner_val.name}"
                    group_parser = make_argument(
                        arg_name, List[inner_val.type], group_parser
                    )
            # If it's a reference to a class it needs to be an arg of a simple string as class matching will take care
            # of it later on
            elif val_type.__module__ == "spock.backend.config":
                arg_name = f"--{str(attr_name)}.{val.name}"
                val_type = str
                group_parser = make_argument(arg_name, val_type, group_parser)
            # This catches callables -- need to be of type str which will be use in importlib
            elif isinstance(val.type, _SpockVariadicGenericAlias):
                arg_name = f"--{str(attr_name)}.{val.name}"
                group_parser = make_argument(arg_name, str, group_parser)
            else:
                arg_name = f"--{str(attr_name)}.{val.name}"
                group_parser = make_argument(arg_name, val_type, group_parser)
        return parser

    def _extract_fnc(self, val, module_name):
        """Function that gets the nested lists within classes

        Args:
            val: current attr
            module_name: matching module name

        Returns:
            list of any nested classes/enums

        """
        return (
            self._extract_other_types(val.metadata["type"], module_name)
            if "type" in val.metadata
            else []
        )
